"""
Hypy is copyright (c) 2009 Cory Dodt.

GENERATED FILE -- DO NOT EDIT
"""
from datetime import date

author = "Cory Dodt"
releaseDate = "2014-07-27"
releaseRevision = "b480ff576daa tip"
__version__ = "0.8.5.1"
