"""Hypy - Pythonic Hyper Estraier"""
from hypy.lib import (PutFailed,
    OpenFailed,
    CloseFailed,
    EditFailed,
    FlushFailed,
    HCondition,
    HDatabase,
    HResults,
    HDocument,
    HHit,
    )

from hypy.copyright import __version__
